@@ -19,7 +19,7 @@ Boston, MA 02111-1307, USA.
 *-GNU-GPL-END-*/
 
 /*
-  $Id: ruli_addr.h,v 1.4 2004/06/01 18:37:17 evertonm Exp $
+  $Id: ruli_addr.h,v 1.6 2004/06/16 23:05:58 evertonm Exp $
   */
 
 
@@ -61,7 +61,11 @@ struct in6_addr ruli_addr_inet6(const ruli_addr_t *addr);
 int ruli_inet6_aton(const char *cp, struct in6_addr *inp);
 int ruli_addr_parse(const char *p, ruli_addr_t *addr);
 ruli_addr_t *ruli_addr_parse_new(const char *p);
+int ruli_inet6_printf(FILE *out, const char *fmt, const struct in6_addr *ia);
+int ruli_inet6_print(FILE *out, const struct in6_addr *ia);
 int ruli_addr_print(FILE *out, const ruli_addr_t *addr);
+int ruli_inet6_snprintf(char *buf, size_t size, const char *fmt, const struct in6_addr *ia);
+int ruli_inet6_snprint(char *buf, size_t size, const struct in6_addr *ia);
 int ruli_addr_snprint(char *buf, size_t size, const ruli_addr_t *addr);
 
 #endif /* RULI_ADDR_H */
