@@ -19,7 +19,7 @@ Boston, MA 02111-1307, USA.
 *-GNU-GPL-END-*/
 
 /*
-  $Id: ruli_conf.c,v 1.7 2004/06/03 17:34:27 evertonm Exp $
+  $Id: ruli_conf.c,v 1.8 2004/06/16 21:23:11 evertonm Exp $
  */
 
 
@@ -30,6 +30,7 @@ Boston, MA 02111-1307, USA.
 
 #include <stdio.h>
 #include <assert.h>
+#include <strings.h>
 #include <string.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -44,6 +45,12 @@ Boston, MA 02111-1307, USA.
 #define LOAD_NS_LIST_INBUFSZ     1024
 
 
+#ifdef _RULI_SOLARIS_
+/* should be in <string.h> */
+extern char *strtok_r(char *s, const char *delim, char **ptrptr);
+#endif
+
+
 static int conf_load_search_list(ruli_list_t *search_list)
 {
   const char *const RESOLV_CONF = "/etc/resolv.conf";
