@@ -19,7 +19,7 @@ Boston, MA 02111-1307, USA.
 *-GNU-GPL-END-*/
 
 /*
-  $Id: ruli_search.h,v 1.5 2004/06/14 17:05:03 evertonm Exp $
+  $Id: ruli_search.h,v 1.6 2004/06/21 00:22:48 evertonm Exp $
 */
 
 
@@ -56,16 +56,21 @@ typedef struct {
 /*
  * Query functions
  */
-ruli_search_srv_t *_ruli_search_srv_submit(void *(*fallback_call)(ruli_srv_t *srv_query),
+ruli_search_srv_t *_ruli_search_srv_submit(void *(*fallback_call)
+					   (ruli_srv_t *srv_query),
 					   ruli_res_t *resolver, 
-					   void *(*call)(ruli_search_srv_t *search, void *arg),
+					   void *(*call)
+					   (ruli_search_srv_t *search,
+					    void *arg),
 					   void *call_arg,
 					   long options,
 					   const char *txt_service,
 					   const char *txt_domain,
 					   int fallback_port);
 ruli_search_srv_t *ruli_search_srv_submit(ruli_res_t *resolver, 
-					  void *(*call)(ruli_search_srv_t *search, void *arg),
+					  void *(*call)
+					  (ruli_search_srv_t *search, 
+					   void *arg),
 					  void *call_arg,
 					  long options,
 					  const char *txt_service,
