@@ -19,7 +19,7 @@ Boston, MA 02111-1307, USA.
 *-GNU-GPL-END-*/
 
 /*
-  $Id: ruli_srv.c,v 1.71 2004/06/23 22:10:56 evertonm Exp $
+  $Id: ruli_srv.c,v 1.74 2004/06/25 23:19:11 evertonm Exp $
   */
 
 
@@ -68,18 +68,16 @@ static void *query_done(ruli_srv_t *srv_qry, int srv_result_code)
 /*
   dispose fallback query
  */
-static void f_query_done(ruli_srv_t *srv_qry) {
-  assert(srv_qry->fall_query);
-  
-  ruli_res_query_delete(srv_qry->fall_query);
-  ruli_free(srv_qry->fall_query);
-  srv_qry->fall_query = 0;
+static void f_query_done(ruli_host_t *fall_qry) {
+  assert(fall_qry);
+  ruli_host_query_delete(fall_qry);
+  ruli_free(fall_qry);
 }
 
 /*
   finish fallback query
  */
-static void *fall_query_done(ruli_srv_t *srv_qry, int srv_result_code)
+static void *fall_query_done(ruli_srv_t *srv_qry, int srv_result_code, ruli_host_t *fall_qry)
 {
 #ifdef RULI_SRV_DEBUG
   fprintf(stderr, 
@@ -90,28 +88,34 @@ static void *fall_query_done(ruli_srv_t *srv_qry, int srv_result_code)
 	  srv_result_code);
 #endif
 
-  f_query_done(srv_qry);
+  f_query_done(fall_qry);
 
   return query_done(srv_qry, srv_result_code);
 }
 
+/* walk query context */
+typedef struct {
+  ruli_host_t walk_query;  /* auxiliary query context for "walk" */
+  ruli_srv_t  *srv_query;  /* original srv query */
+} walk_t;
+
 /*
   dispose walk query
  */
-static void w_query_done(ruli_srv_t *srv_qry)
+static void w_query_done(walk_t *walk)
 {
-  assert(srv_qry->walk_query);
-
-  ruli_res_query_delete(srv_qry->walk_query);
-  ruli_free(srv_qry->walk_query);
-  srv_qry->walk_query = 0;
+  assert(walk);
+  ruli_host_query_delete(&walk->walk_query);
+  ruli_free(walk);
 }
 
 /*
   finish walk query
  */
-static void *walk_query_done(ruli_srv_t *srv_qry, int srv_result_code)
+static void *walk_query_done(walk_t *walk, int srv_result_code)
 {
+  ruli_srv_t *srv_qry = walk->srv_query;
+
 #ifdef RULI_SRV_DEBUG
   fprintf(stderr, 
 	  "DEBUG: %s: %s(): %d: id=%d query_result=%s [%d]\n", 
@@ -121,31 +125,23 @@ static void *walk_query_done(ruli_srv_t *srv_qry, int srv_result_code)
 	  srv_result_code);
 #endif
 
-  w_query_done(srv_qry);
+  w_query_done(walk);
 
   return query_done(srv_qry, srv_result_code);
 }
 
-static void *on_walk_answer(ruli_res_query_t *qry, void *qry_arg)
+static void *on_walk_answer(ruli_host_t *host_qry, void *qry_arg)
 {
-  ruli_srv_t       *srv_qry = (ruli_srv_t *) qry_arg;
-
+  walk_t           *walk_qry = (walk_t *) qry_arg;
+  ruli_srv_t       *srv_qry = walk_qry->srv_query;
   ruli_srv_entry_t *srv_entry;
-  ruli_list_t      *addr_list;
   int              prev_addr_list_size;
 
-  assert(qry->answer_code != RULI_SRV_CODE_VOID);
-  assert(srv_qry->walk_query);
-  assert(srv_qry->walk_query == qry);
+  assert(ruli_host_answer_code(host_qry) != RULI_SRV_CODE_VOID);
 
   srv_entry = (ruli_srv_entry_t *) ruli_list_get(&srv_qry->answer_srv_list, 
-			      	                 srv_qry->walk_index);
-  assert(srv_entry);
-
-  addr_list           = &srv_entry->addr_list;
-  assert(addr_list);
-
-  prev_addr_list_size = ruli_list_size(addr_list);
+                                                 srv_qry->under.walk_index);
+  prev_addr_list_size = ruli_list_size(&srv_entry->addr_list);
 
 #ifdef RULI_SRV_DEBUG
   {
@@ -168,219 +164,85 @@ static void *on_walk_answer(ruli_res_query_t *qry, void *qry_arg)
   /*
    * Query failed?
    */
-  if (qry->answer_code) {
-    srv_qry->w_query_rcode = RULI_RCODE_VOID;
+  {
+    int answer_code = ruli_host_answer_code(host_qry);
+    if (answer_code) {
+      srv_qry->last_rcode = RULI_RCODE_VOID;
 
-    if (qry->answer_code == RULI_CODE_TIMEOUT)
-      return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_ALARM);
+      switch(answer_code) {
+      case RULI_HOST_CODE_ALARM:
+        return walk_query_done(walk_qry, RULI_SRV_CODE_WALK_ALARM);
+      case RULI_HOST_CODE_EMPTY:
+        return walk_query_done(walk_qry, RULI_SRV_CODE_WALK_EMPTY);
+      default:
+        return walk_query_done(walk_qry, RULI_SRV_CODE_WALK_QUERY);
+      }
 
-    return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_QUERY);
+      assert(0);
+    }
   }
 
   /*
    * Bad RCODE ?
    */
   {
-    ruli_uint16_t rcode = qry->answer_header.rcode;
+    ruli_uint16_t rcode = ruli_host_rcode(host_qry);
+    srv_qry->last_rcode = rcode;
+    if (rcode != RULI_RCODE_NOERROR)
+      return walk_query_done(walk_qry, RULI_SRV_CODE_FALL_RCODE);
+  }
 
-    srv_qry->w_query_rcode = rcode;
+  /*
+   * Move addresses into SRV record
+   */
+  {
+    ruli_list_t *addr_list = &host_qry->answer_addr_list;
+    int addr_list_size = ruli_list_size(addr_list);
+    int i;
 
-    if (rcode != RULI_RCODE_NOERROR) {
-
-      if (_ruli_need_query_in_a(srv_qry->walk_qtype, srv_qry->srv_options)) {
-	/* Dispose auxiliary query as it will be re-created
-	   by next walk query */
-	w_query_done(srv_qry);
-
-	srv_qry->walk_qtype = RULI_RR_TYPE_A;
-
-	return _ruli_srv_answer_walk(srv_qry); /* repeat walk index */
+    /* Move addresses from addr query to SRV entry */
+    for (i = 0; i < addr_list_size; ++i) {
+      ruli_addr_t *addr = ruli_list_get(addr_list, i);
+      if (ruli_list_push(&srv_entry->addr_list, addr)) {
+        ruli_list_prune(&srv_entry->addr_list, 0);
+        return walk_query_done(walk_qry, RULI_SRV_CODE_FALL_OTHER);
       }
-
-      return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_RCODE);
     }
   }
 
   /*
-   * Parse answer
+   * Detaches addresses from addr-query ruli_list_t container
+   * So that next ruli_host_query_delete does not dispose them
    */
-  {
-    ruli_parse_t parse;
-
-    if (ruli_parse_new(&parse))
-      return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_OTHER);
-
-    {
-      int result;
-
-      assert(sizeof(ruli_uint8_t) == sizeof(char));
-
-      result = ruli_parse_message(&parse, &qry->answer_header,
-			          (ruli_uint8_t *) qry->answer_buf,
-			          qry->answer_msg_len);
-      if (result) {
-	ruli_parse_delete(&parse);
-	return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_PARSE);
-      }
-    }
-
-    /*
-     * As RFC2782 states alias aren't valid SRV targets,
-     * we only consider address records in the answer.
-     * If alias were permitted, we would need to
-     * consider IN CNAME RRs.
-     */
-
-    /*
-     * Search for address records in answer
-     */
-    {
-      ruli_list_t *an_list     = &parse.answer_list;
-      int         an_list_size = ruli_list_size(an_list);
-      int         i;
-
-      for (i = 0; i < an_list_size; ++i) {
-	ruli_rr_t *rr = (ruli_rr_t *) ruli_list_get(an_list, i);
-
-	if (rr->qclass != RULI_RR_CLASS_IN)
-	  continue;
-	if (!ruli_rr_type_is_address(srv_qry->srv_options, rr->type))
-	  continue;
-
-	/*
-	 * Owner matches?
-	 */
-
-        assert(sizeof(ruli_uint8_t) == sizeof(char));
-
-        if (ruli_dname_compare(rr->owner,
-                               (ruli_uint8_t *) qry->answer_buf,
-                               qry->answer_msg_len,
-                               (ruli_uint8_t *) qry->full_dname,
-                               qry->full_dname_len))
-          continue;
-
-	/*
-	 * Save address
-	 */
-	{
-	  /* Allocate space */
-	  ruli_addr_t *addr = (ruli_addr_t *) ruli_malloc(sizeof(*addr));
-	  if (!addr) {
-	    ruli_parse_delete(&parse);
-	    return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_OTHER);
-	  }
-
-	  /* Save space */
-	  if (ruli_list_push(addr_list, addr)) {
-	    ruli_free(addr); 
-	    ruli_parse_delete(&parse);
-	    return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_OTHER);
-	  }
-
-	  /* Write address into space */
-	  ruli_parse_addr_rr(addr, rr, srv_qry->srv_options);
-
-	} /* save address */
-
-      } /* for */
-    }
-
-    ruli_parse_delete(&parse);
-  }
-
-  /* Require at least one address has been added */
-  {
-    int curr_addr_list_size = ruli_list_size(addr_list);
-
-    assert(curr_addr_list_size >= prev_addr_list_size);
-
-    if (curr_addr_list_size == prev_addr_list_size) {
-#ifdef RULI_SRV_DEBUG
-      {
-	char target_txt[RULI_LIMIT_DNAME_TEXT_BUFSZ];
-	int  target_txt_len;
-	int  result;
-	
-	result = ruli_dname_decode(target_txt, RULI_LIMIT_DNAME_TEXT_BUFSZ, 
-				   &target_txt_len, 
-				   (const char *) qry->full_dname, 
-				   qry->full_dname_len);
-	assert(!result);
-	
-	fprintf(stderr, 
-		"DEBUG: %s: %s(): %d: "
-		"target=%s len=%d w_qtype=%d w_index=%d: "
-		"BUT: no matching address record\n", 
-		__FILE__, __PRETTY_FUNCTION__, __LINE__,
-		target_txt, target_txt_len, 
-		srv_qry->walk_qtype, srv_qry->walk_index);
-      }
-#endif
-
-      if (_ruli_need_query_in_a(srv_qry->walk_qtype, srv_qry->srv_options)) {
-	/* Dispose auxiliary query as it will be re-created
-	   by next walk query */
-	w_query_done(srv_qry);
-
-	srv_qry->walk_qtype = RULI_RR_TYPE_A;
-
-	return _ruli_srv_answer_walk(srv_qry); /* repeat walk index */
-      }
-
-      return walk_query_done(srv_qry, RULI_SRV_CODE_WALK_EMPTY);
-    }
-  }
-
-  if (_ruli_need_query_in_a(srv_qry->walk_qtype, srv_qry->srv_options)) {
-    /* Dispose auxiliary query as it will be re-created
-       by next walk query */
-    w_query_done(srv_qry);
-
-    srv_qry->walk_qtype = RULI_RR_TYPE_A;
-
-    return _ruli_srv_answer_walk(srv_qry); /* repeat walk index */
-  }
+  ruli_host_query_detach(host_qry);
 
   /* Dispose auxiliary query as it will be re-created
      by next walk query */
-  w_query_done(srv_qry);
-
-  /* For the next target, use the first address type */
-  srv_qry->walk_qtype = _ruli_reset_qtype(srv_qry->srv_options);
+  w_query_done(walk_qry);
 
   /* In future walk query, consider next answer target */
-  ++srv_qry->walk_index;
+  ++srv_qry->under.walk_index;
 
   return _ruli_srv_answer_walk(srv_qry);
 }
 
-static int find_addr_type(const ruli_list_t *addr_list, int qtype)
+static int find_addr(const ruli_list_t *addr_list, long options)
 {
   int list_size = ruli_list_size(addr_list);
   int i;
 
-  switch (qtype) {
-  case RULI_RR_TYPE_A:
-
-    for (i = 0; i < list_size; ++i) {
-      ruli_addr_t *addr = (ruli_addr_t *) ruli_list_get(addr_list, i);
-      if (ruli_addr_family(addr) == PF_INET)
-	return -1;
+  for (i = 0; i < list_size; ++i) {
+    ruli_addr_t *addr = (ruli_addr_t *) ruli_list_get(addr_list, i);
+    switch (ruli_addr_family(addr)) {
+      case PF_INET6:
+        if (!(options & RULI_RES_OPT_SRV_NOINET6))
+          return -1;
+        break;
+      case PF_INET:
+        if (!(options & RULI_RES_OPT_SRV_NOINET))
+          return -1;
+        break;
     }
-    break;
-
-  case RULI_RR_TYPE_AAAA:
-
-    for (i = 0; i < list_size; ++i) {
-      ruli_addr_t *addr = (ruli_addr_t *) ruli_list_get(addr_list, i);
-      if (ruli_addr_family(addr) == PF_INET6)
-	return -1;
-    }
-    break;
-
-  default:
-    assert(0);
   }
 
   return 0;
@@ -402,15 +264,16 @@ void *_ruli_srv_answer_walk(ruli_srv_t *srv_qry)
     return query_done(srv_qry, RULI_SRV_CODE_OK);
 
   /*
-   * Scan answer targets, considering address lists
+   * Scan SRV answer targets, considering address lists
    */
-  for (; srv_qry->walk_index < srv_list_size; ++srv_qry->walk_index) {
+  for (; srv_qry->under.walk_index < srv_list_size; ++srv_qry->under.walk_index) {
     ruli_srv_entry_t *entry = (ruli_srv_entry_t *) ruli_list_get(srv_list, 
-						   srv_qry->walk_index);
+						   srv_qry->under.walk_index);
     ruli_list_t *addr_list = &entry->addr_list;
+    walk_t *walk_qry;
 
     /* If this target already has address(es), skip it */
-    if (find_addr_type(addr_list, srv_qry->walk_qtype))
+    if (find_addr(addr_list, srv_qry->srv_options))
       continue;
 
 #ifdef RULI_SRV_DEBUG
@@ -434,32 +297,29 @@ void *_ruli_srv_answer_walk(ruli_srv_t *srv_qry)
     /*
      * Allocate space for auxiliary walk query
      */
-    assert(!srv_qry->walk_query);
-    srv_qry->walk_query = \
-      (ruli_res_query_t *) ruli_malloc(sizeof(ruli_res_query_t));
-    if (!srv_qry->walk_query)
+    walk_qry = \
+      (walk_t *) ruli_malloc(sizeof(*walk_qry));
+    if (!walk_qry)
       return query_done(srv_qry, RULI_SRV_CODE_WALK_OTHER);
-
-    assert( (srv_qry->walk_qtype == RULI_RR_TYPE_A) ||
-	    (srv_qry->walk_qtype == RULI_RR_TYPE_AAAA) );
+    walk_qry->srv_query = srv_qry;
 
     /*
      * Initialize walk query arguments
      */
-    srv_qry->walk_query->q_on_answer     = on_walk_answer;
-    srv_qry->walk_query->q_on_answer_arg = srv_qry;
-    srv_qry->walk_query->q_domain        = entry->target;
-    srv_qry->walk_query->q_domain_len    = entry->target_len;
-    srv_qry->walk_query->q_class         = RULI_RR_CLASS_IN;
-    srv_qry->walk_query->q_type          = srv_qry->walk_qtype;
-    srv_qry->walk_query->q_options       = RULI_RES_OPT_VOID;
+    walk_qry->walk_query.host_resolver        = srv_qry->srv_resolver;
+    walk_qry->walk_query.host_on_answer       = on_walk_answer;
+    walk_qry->walk_query.host_on_answer_arg   = walk_qry;
+    walk_qry->walk_query.host_domain          = entry->target;
+    walk_qry->walk_query.host_domain_len      = entry->target_len;
+    walk_qry->walk_query.host_options         = srv_qry->srv_options;
+    /* RFC 2782 states CNAME aren't valid SRV targets */
+    walk_qry->walk_query.host_max_cname_depth = 0;
 
     /*
      * Submit walk query
      */
-    if (ruli_res_query_submit(srv_qry->srv_resolver, srv_qry->walk_query)) {
-      ruli_free(srv_qry->walk_query);
-      srv_qry->walk_query = 0;
+    if (ruli_host_query_submit(&walk_qry->walk_query)) {
+      ruli_free(walk_qry);
       return query_done(srv_qry, RULI_SRV_CODE_WALK_QUERY);
     }
 
@@ -475,381 +335,81 @@ void *_ruli_srv_answer_walk(ruli_srv_t *srv_qry)
   return query_done(srv_qry, RULI_SRV_CODE_OK);
 }
 
-static void *fallback_resubmit(ruli_srv_t *srv_qry)
-{
-#ifdef RULI_SRV_DEBUG
-  fprintf(stderr, 
-	  "DEBUG: %s: %s(): %d: OLD query_id=%d\n", 
-	  __FILE__, __PRETTY_FUNCTION__, __LINE__,
-	  srv_qry->fall_query->query_id);
-#endif /* RULI_SRV_DEBUG */
-
-  /* reset fallback query, but not dispose it */
-  assert(srv_qry->fall_query);
-  ruli_res_query_delete(srv_qry->fall_query);
-
-  srv_qry->fall_query->q_type = srv_qry->fall_qtype;
-
-  assert(!_ruli_need_query_in_a(srv_qry->fall_qtype, srv_qry->srv_options));
-
-  if (ruli_res_query_submit(srv_qry->srv_resolver, srv_qry->fall_query)) {
-    ruli_free(srv_qry->fall_query);
-    srv_qry->fall_query = 0;
-    return query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER);
-  }
-
-#ifdef RULI_SRV_DEBUG
-  fprintf(stderr, 
-	  "DEBUG: %s: %s(): %d: NEW query_id=%d\n", 
-	  __FILE__, __PRETTY_FUNCTION__, __LINE__,
-	  srv_qry->fall_query->query_id);
-#endif /* RULI_SRV_DEBUG */
-
-  return OOP_CONTINUE;
-}
-
-static void *on_fallback_answer(ruli_res_query_t *qry, void *qry_arg)
+static void *on_fallback_answer(ruli_host_t *qry, void *qry_arg)
 {
   ruli_srv_t *srv_qry           = (ruli_srv_t *) qry_arg;
   int        prev_srv_list_size = ruli_list_size(&srv_qry->answer_srv_list);
+  int        answer_code;
 
-  /* fallback query adds 1 srv entry at most */
-  assert(prev_srv_list_size >= 0);
-  assert(prev_srv_list_size <= 1);
-
-#ifdef RULI_SRV_DEBUG
-  {
-    char txt_dname_buf[RULI_LIMIT_DNAME_TEXT_BUFSZ];
-    int  txt_dname_len;
-    int  result;
-
-    result = ruli_dname_decode(txt_dname_buf, RULI_LIMIT_DNAME_TEXT_BUFSZ,
-			       &txt_dname_len, 
-			       (const char *) qry->full_dname, 
-			       qry->full_dname_len);
-    assert(!result);
-
-    fprintf(stderr, 
-	    "DEBUG: %s: %s(): %d: domain=%s len=%d f_qtype=%d\n",
-	    __FILE__, __PRETTY_FUNCTION__, __LINE__,
-	    txt_dname_buf, txt_dname_len, srv_qry->fall_qtype);
-
-    fprintf(stderr, 
-	    "DEBUG: %s: %s(): %d: id=%d answer_code=%d rcode=%d\n",
-	    __FILE__, __PRETTY_FUNCTION__, __LINE__,
-	    qry->query_id, qry->answer_code, qry->answer_header.rcode);
-  }
-#endif
+  assert(prev_srv_list_size == 0);
 
   assert(qry->answer_code != RULI_SRV_CODE_VOID);
-  assert(srv_qry->fall_query);
-  assert(srv_qry->fall_query == qry);
 
   /*
    * Query failed?
    */
-  if (qry->answer_code) {
-    srv_qry->f_query_rcode = RULI_RCODE_VOID;
+  answer_code = ruli_host_answer_code(qry);
+  if (answer_code) {
+    srv_qry->last_rcode = RULI_RCODE_VOID;
 
-    if (qry->answer_code == RULI_CODE_TIMEOUT)
-      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_ALARM);
+    switch(answer_code) {
+    case RULI_HOST_CODE_ALARM:
+      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_ALARM, qry);
+    case RULI_HOST_CODE_EMPTY:
+      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_EMPTY, qry);
+    default: 
+      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_QUERY, qry);
+    }
 
-    return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_QUERY);
+    assert(0);
   }
 
   /*
    * Bad RCODE ?
    */
   {
-    ruli_uint16_t rcode = qry->answer_header.rcode;
-
-    srv_qry->f_query_rcode = rcode;
-
-    if (rcode != RULI_RCODE_NOERROR) {
-
-      if (_ruli_need_query_in_a(srv_qry->fall_qtype, srv_qry->srv_options)) {
-	srv_qry->fall_qtype = RULI_RR_TYPE_A;
-	return fallback_resubmit(srv_qry);
-      }
-
-      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_RCODE);
-    }
+    ruli_uint16_t rcode = ruli_host_rcode(qry);
+    srv_qry->last_rcode = rcode;
+    if (rcode != RULI_RCODE_NOERROR)
+      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_RCODE, qry);
   }
 
   /*
-   * Parse answer
+   * Move addresses into SRV record
    */
   {
-    ruli_uint8_t       *wanted_owner;
-    int                wanted_owner_len;
-    ruli_cname_rdata_t cname_rdata;
-    int                cname_found; /* boolean */
-    int                remaining_cname_depth = RULI_LIMIT_CNAME_DEPTH;
+    ruli_srv_entry_t *srv_entry;
+    ruli_list_t *addr_list = &qry->answer_addr_list;
+    int addr_list_size = ruli_list_size(addr_list);
+    int i;
 
-    ruli_parse_t parse;
+    /* Create SRV entry and append it to list */
+    srv_entry =_ruli_srv_list_new_entry(&srv_qry->answer_srv_list,
+                                        srv_qry->srv_domain,
+                                        srv_qry->srv_domain_len,
+                                        -1,
+                                        -1,
+                                        srv_qry->srv_fallback_port);
+    if (!srv_entry)
+      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER, qry);
 
-    if (ruli_parse_new(&parse))
-      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER);
-
-    {
-      int result;
-
-      assert(sizeof(ruli_uint8_t) == sizeof(char));
-
-      result = ruli_parse_message(&parse, &qry->answer_header,
-				  (ruli_uint8_t *) qry->answer_buf,
-                                  qry->answer_msg_len);
-      if (result) {
-	ruli_parse_delete(&parse);
-	return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_PARSE);
+    /* Move addresses from addr query to SRV entry */
+    for (i = 0; i < addr_list_size; ++i) {
+      ruli_addr_t *addr = ruli_list_get(addr_list, i);
+      if (ruli_list_push(&srv_entry->addr_list, addr)) {
+        ruli_list_prune(&srv_entry->addr_list, 0);
+        return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER, qry);
       }
     }
-
-    assert(sizeof(ruli_uint8_t) == sizeof(char));
-
-    wanted_owner     = (ruli_uint8_t *) qry->full_dname;
-    wanted_owner_len = qry->full_dname_len;
-
-    /*
-     * Search for Address records
-     *
-     * The following algorithm scans the answer section
-     * for CNAME records:
-     *
-     * 1/3: We look for address records for the wanted_owner
-     * 2/3: If at least one address record is found, we're done
-     * 3/3: We search IN CNAME records for the wanted_owner
-     *      If a CNAME record is found, we make
-     *      wanted_owner <- cname_target and go to 1/3
-     *      Otherwise, if no CNAME found, we quit
-     */
-    for (;;) {
-      ruli_list_t      *an_list     = &parse.answer_list;
-      int              an_list_size = ruli_list_size(an_list);
-      ruli_srv_entry_t *srv_entry   = prev_srv_list_size ? \
-	(ruli_srv_entry_t *) ruli_list_top(&srv_qry->answer_srv_list) : 0;
-      int              i;
-
-#ifdef RULI_SRV_DEBUG
-      {
-	char wanted_txt[RULI_LIMIT_DNAME_TEXT_BUFSZ];
-	int  wanted_txt_len;
-	int  result;
-	
-	assert(sizeof(ruli_uint8_t) == sizeof(char));
-	
-	result = ruli_dname_decode(wanted_txt, RULI_LIMIT_DNAME_TEXT_BUFSZ, 
-				   &wanted_txt_len, 
-				   (const char *) wanted_owner, 
-				   wanted_owner_len);
-	assert(!result);
-	
-	fprintf(stderr, 
-		"DEBUG: %s: %s(): %d: "
-		"wanted owner=(%d)%s answer_list_len=%d\n", 
-		__FILE__, __PRETTY_FUNCTION__, __LINE__,
-		wanted_txt_len, wanted_txt, an_list_size);
-      }
-#endif
-
-      /*
-       * 1/3: Scan answer section for address records
-       */
-      for (i = 0; i < an_list_size; ++i) {
-	ruli_rr_t *rr = (ruli_rr_t *) ruli_list_get(an_list, i);
-
-	if (rr->qclass != RULI_RR_CLASS_IN)
-	  continue;
-	if (!ruli_rr_type_is_address(srv_qry->srv_options, rr->type))
-	  continue;
-
-	/*
-	 * Owner matches?
-	 */
-
-        if (ruli_dname_compare(rr->owner,
-                               (ruli_uint8_t *) qry->answer_buf,
-                               qry->answer_msg_len,
-                               wanted_owner,
-                               wanted_owner_len))
-          continue;
-
-	/*
-	 * Save address
-	 */
-	{
-	  ruli_addr_t *addr;
-
-	  if (!srv_entry) {
-
-	    /* 
-	     * Create SRV record and append it to list 
-	     */
-	    srv_entry =_ruli_srv_list_new_entry(&srv_qry->answer_srv_list,
-						srv_qry->srv_domain,
-						srv_qry->srv_domain_len,
-						-1,
-						-1,
-						srv_qry->srv_fallback_port);
-	    if (!srv_entry) {
-	      ruli_parse_delete(&parse);
-	      return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER);
-	    }
-	    
-	  } /* Create SRV record */
-
-	  /* Allocate space */
-	  addr = (ruli_addr_t *) ruli_malloc(sizeof(*addr));
-	  if (!addr) {
-	    ruli_parse_delete(&parse);
-	    return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER);
-	  }
-
-	  /* Save space */
-	  if (ruli_list_push(&srv_entry->addr_list, addr)) {
-	    ruli_free(addr); 
-	    ruli_parse_delete(&parse);
-	    return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER);
-	  }
-
-	  /* Write address into space */
-	  ruli_parse_addr_rr(addr, rr, srv_qry->srv_options);
-
-	} /* save address */
-
-      } /* for: scan answer section for addresses */
-
-      /* 
-       * 2/3: Done if one SRV record has been built 
-       */
-      {
-	int curr_srv_list_size = ruli_list_size(&srv_qry->answer_srv_list);
-	
-	assert(curr_srv_list_size >= 0);
-	assert(curr_srv_list_size <= 1);
-	
-	if (curr_srv_list_size > 0) {
-	  ruli_parse_delete(&parse);
-
-	  if (_ruli_need_query_in_a(srv_qry->fall_qtype, srv_qry->srv_options)) {
-	    srv_qry->fall_qtype = RULI_RR_TYPE_A;
-	    return fallback_resubmit(srv_qry);
-	  }
-
-	  return fall_query_done(srv_qry, RULI_SRV_CODE_OK);
-	}
-      }
-
-#ifdef RULI_SRV_DEBUG
-      fprintf(stderr, 
-	      "DEBUG: %s: %s(): %d: id=%d answer_code=%d, rcode=%d: "
-	      "BUT: no matching address record\n",
-	      __FILE__, __PRETTY_FUNCTION__, __LINE__,
-	      qry->query_id, qry->answer_code, qry->answer_header.rcode);
-#endif
-
-      /*
-       * 3/3: Scan answer section for IN CNAME records
-       */
-
-      cname_found = 0; /* false */
-
-      for (i = 0; i < an_list_size; ++i) {
-	ruli_rr_t *rr = (ruli_rr_t *) ruli_list_get(an_list, i);
-
-	if (rr->qclass != RULI_RR_CLASS_IN)
-	  continue;
-	if (rr->type != RULI_RR_TYPE_CNAME)
-	  continue;
-
-	/*
-	 * Owner matches?
-	 */
-
-        if (ruli_dname_compare(rr->owner,
-                               (ruli_uint8_t *) qry->answer_buf,
-                               qry->answer_msg_len,
-                               wanted_owner,
-                               wanted_owner_len))
-	  continue;
-	
-	/* Get new owner from CNAME target */
-	
-	assert(sizeof(ruli_uint8_t) == sizeof(char));
-	
-	if (ruli_parse_rr_cname(&cname_rdata, rr->rdata, rr->rdlength,
-				(ruli_uint8_t *) qry->answer_buf,
-				qry->answer_msg_len)) {
-	  ruli_parse_delete(&parse);
-	  return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_PARSE);
-	}
-	
-#ifdef RULI_SRV_DEBUG
-	{
-	  ruli_uint8_t cname_owner_txt[RULI_LIMIT_DNAME_TEXT_BUFSZ];
-	  int          cname_owner_txt_len;
-	  char         cname_trg_txt[RULI_LIMIT_DNAME_TEXT_BUFSZ];
-	  int          cname_trg_txt_len;
-	  int          result;
-	  
-	  assert(sizeof(ruli_uint8_t) == sizeof(char));
-	  
-	  result = ruli_dname_extract((const ruli_uint8_t *) qry->answer_buf,
-				     (const ruli_uint8_t *) qry->answer_buf + qry->answer_msg_len,
-				     cname_owner_txt, 
-				     cname_owner_txt + RULI_LIMIT_DNAME_TEXT_BUFSZ,
-				     rr->owner, &cname_owner_txt_len); 
-	  assert(!result);
-	  
-	  result = ruli_dname_decode(cname_trg_txt, RULI_LIMIT_DNAME_TEXT_BUFSZ, 
-				     &cname_trg_txt_len, 
-				     (const char *) cname_rdata.cname, 
-				     cname_rdata.cname_len);
-	  assert(!result);
-	  
-	  fprintf(stderr, 
-		  "DEBUG: %s: %s(): %d: "
-		  "CNAME owner=(%d)%s target=(%d)%s remaining=%d\n", 
-		  __FILE__, __PRETTY_FUNCTION__, __LINE__,
-		  cname_owner_txt_len, cname_owner_txt,
-		  cname_trg_txt_len, cname_trg_txt,
-		  remaining_cname_depth);
-	}
-#endif
-	
-	wanted_owner     = cname_rdata.cname;
-	wanted_owner_len = cname_rdata.cname_len;
-	
-	cname_found = 1; /* true */
-	
-	break;
-
-      } /* for: IN CNAME scan of answer section */
-
-      /* If no CNAME found, give up */
-      if (!cname_found)
-	break;
-
-      /* If CNAME found, check depth */
-      if (--remaining_cname_depth < 0) {
-	ruli_parse_delete(&parse);
-	return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_CNAME);
-      }
-
-      /* Loop to find address for CNAME target */
-
-    } /* for: Address search */    
-
-    ruli_parse_delete(&parse);
-    
-  } /* parsed answer context */
-  
-  if (_ruli_need_query_in_a(srv_qry->fall_qtype, srv_qry->srv_options)) {
-    srv_qry->fall_qtype = RULI_RR_TYPE_A;
-    return fallback_resubmit(srv_qry);
   }
 
-  return fall_query_done(srv_qry, RULI_SRV_CODE_FALL_EMPTY);
+  /* 
+   * Detaches addresses from addr-query ruli_list_t container
+   * So that next ruli_host_query_delete does not dispose them
+   */
+  ruli_host_query_detach(qry);
+
+  return fall_query_done(srv_qry, RULI_SRV_CODE_OK, qry);
 }
 
 /*
@@ -857,6 +417,8 @@ static void *on_fallback_answer(ruli_res_query_t *qry, void *qry_arg)
  */
 void *_ruli_srv_answer_fallback_addr(ruli_srv_t *srv_qry)
 {
+  ruli_host_t *fall_qry;
+
 #ifdef RULI_SRV_DEBUG
   {
     char txt_dname_buf[RULI_LIMIT_DNAME_TEXT_BUFSZ];
@@ -877,9 +439,6 @@ void *_ruli_srv_answer_fallback_addr(ruli_srv_t *srv_qry)
   }
 #endif
 
-  assert(!srv_qry->fall_query);
-  assert(srv_qry->fall_qtype == -1);
-
   /* Has the user disabled fallback query? */
   if (srv_qry->srv_options & RULI_RES_OPT_SRV_NOFALL)
     return query_done(srv_qry, RULI_SRV_CODE_OK);
@@ -887,30 +446,27 @@ void *_ruli_srv_answer_fallback_addr(ruli_srv_t *srv_qry)
   /*
    * Allocate space for fallback query
    */
-  srv_qry->fall_query = (ruli_res_query_t *) \
-    ruli_malloc(sizeof(ruli_res_query_t));
-  if (!srv_qry->fall_query)
+  fall_qry = (ruli_host_t *) \
+    ruli_malloc(sizeof(*fall_qry));
+  if (!fall_qry)
     return query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER);
 
-  srv_qry->fall_qtype = _ruli_reset_qtype(srv_qry->srv_options);
-
   /*
    * Initialize fallback query arguments
    */
-  srv_qry->fall_query->q_on_answer     = on_fallback_answer;
-  srv_qry->fall_query->q_on_answer_arg = srv_qry;
-  srv_qry->fall_query->q_domain        = srv_qry->srv_domain;
-  srv_qry->fall_query->q_domain_len    = srv_qry->srv_domain_len;
-  srv_qry->fall_query->q_class         = RULI_RR_CLASS_IN;
-  srv_qry->fall_query->q_type          = srv_qry->fall_qtype;
-  srv_qry->fall_query->q_options       = srv_qry->srv_options;
+  fall_qry->host_resolver        = srv_qry->srv_resolver;
+  fall_qry->host_on_answer       = on_fallback_answer;
+  fall_qry->host_on_answer_arg   = srv_qry;
+  fall_qry->host_domain          = srv_qry->srv_domain;
+  fall_qry->host_domain_len      = srv_qry->srv_domain_len;
+  fall_qry->host_options         = srv_qry->srv_options;
+  fall_qry->host_max_cname_depth = RULI_LIMIT_CNAME_DEPTH;
 
   /*
    * Submit fallback query
    */
-  if (ruli_res_query_submit(srv_qry->srv_resolver, srv_qry->fall_query)) {
-    ruli_free(srv_qry->fall_query);
-    srv_qry->fall_query = 0;
+  if (ruli_host_query_submit(fall_qry)) {
+    ruli_free(fall_qry);
     return query_done(srv_qry, RULI_SRV_CODE_FALL_OTHER);
   }
 
@@ -1455,12 +1011,8 @@ static void *on_srv_answer(ruli_res_query_t *qry, void *arg)
    * Step 6/6: Launch queries to fill missing addresses, if any
    */
 
-  assert(!srv_qry->walk_query);
-  assert(srv_qry->walk_index == -1);
-  assert(srv_qry->walk_qtype == -1);
-
-  srv_qry->walk_index = 0;
-  srv_qry->walk_qtype = _ruli_reset_qtype(srv_qry->srv_options);
+  assert(srv_qry->under.walk_index == -1);
+  srv_qry->under.walk_index = 0;
 
   return _ruli_srv_answer_walk(srv_qry);
 }
@@ -1570,13 +1122,9 @@ int _ruli_srv_query_submit(ruli_srv_t *srv_qry,
   }
 
   srv_qry->answer_code = RULI_SRV_CODE_VOID;
+  srv_qry->last_rcode = RULI_RCODE_VOID;
 
-  srv_qry->walk_index = -1;
-  srv_qry->walk_qtype = -1;
-  srv_qry->walk_query = 0;
-
-  srv_qry->fall_qtype = -1;
-  srv_qry->fall_query = 0;
+  srv_qry->under.walk_index = -1;
 
   /*
    * Define callback parameters
@@ -1659,12 +1207,6 @@ void ruli_srv_query_delete(ruli_srv_t *srv_qry)
   }
   ruli_list_dispose_trivial(&srv_qry->answer_srv_list);
 
-  if (srv_qry->fall_query)
-    f_query_done(srv_qry);
-
-  if (srv_qry->walk_query)
-    w_query_done(srv_qry);
-
   ruli_parse_delete(&srv_qry->parse);
   ruli_list_dispose_trivial(&srv_qry->rr_srv_list);
   ruli_list_delete(&srv_qry->pri_srv_list);
@@ -1708,11 +1250,8 @@ int ruli_srv_rcode(ruli_srv_t *srv_qry)
     break;
 
   case RULI_SRV_RCODE_WALK:
-    rcode = srv_qry->w_query_rcode;
-    break;
-
   case RULI_SRV_RCODE_FALL:
-    rcode = srv_qry->f_query_rcode;
+    rcode = srv_qry->last_rcode;
     break;
 
   default:
